class Template:

  template_model = dict(

    Birthdate = dict(
        pid = None,
        process_type = "http://purl.obolibrary.org/obo/NCIT_C142470",
        attribute_type = "http://purl.obolibrary.org/obo/NCIT_C68615",
        agent_type = None,
        agent_id = None,
        input_type = None,
        #input_id = None,
        target_type = None,
        target_id = None,     
        output_type = "http://purl.obolibrary.org/obo/NCIT_C70856", 
        output_id = None,
        unit_type = None,
        specific_process_type = None,
        protocol_type = None,        
        protocol_id = None, 
        substance_type = None,      
        substance_id = None,
        activity_type = None,
        frequency_type = None,
        frequency_value = None,        
        concentration_type = None,
        concentration_value = None,
        concentration_unit_type = None,
        value_date = None,
        value_integer = None,
        value_string = None,
        value_float = None,
        value_datatype = "xsd:date",
        comments = None,
        startdate = None,
        enddate = None,
        age = None,
        uniqid = None,
        event_id = None,
        value = None,
        valueIRI = None,
        activity = None,
        target = None,
        agent = None,
        input = None,
        unit = None
        ),
    
    Birthyear = dict(
        pid = None,
        process_type = "http://purl.obolibrary.org/obo/NCIT_C142470",
        attribute_type = "http://purl.obolibrary.org/obo/NCIT_C83164",
        agent_type = None,
        agent_id = None,
        input_type = None,
        #input_id = None,
        target_type = None,
        target_id = None,     
        output_type = "http://purl.obolibrary.org/obo/NCIT_C70856", 
        output_id = None,
        unit_type = None,
        specific_process_type = None,
        protocol_type = None,        
        protocol_id = None, 
        substance_type = None,      
        substance_id = None,
        activity_type = None,
        frequency_type = None,
        frequency_value = None,        
        concentration_type = None,
        concentration_value = None,
        concentration_unit_type = None,
        value_date = None,
        value_integer = None,
        value_string = None,
        value_float = None,
        value_datatype = "xsd:integer",
        comments = None,
        startdate = None,
        enddate = None,
        age = None,
        uniqid = None,
        event_id = None,
        value = None,
        valueIRI = None,
        activity = None,
        target = None,
        agent = None,
        input = None,
        unit = None
        ),
        
    Deathdate = dict(
        pid = None,
        process_type = "http://purl.obolibrary.org/obo/NCIT_C142470",
        attribute_type = "http://purl.obolibrary.org/obo/NCIT_C70810",
        agent_type = None,
        agent_id = None,
        input_type = None,
        #input_id = None,
        target_type = None,
        target_id = None,     
        output_type = "http://purl.obolibrary.org/obo/NCIT_C70856", 
        output_id = None,
        unit_type = None,
        specific_process_type = None,
        protocol_type = None,        
        protocol_id = None, 
        substance_type = None,      
        substance_id = None,
        activity_type = None,
        frequency_type = None,
        frequency_value = None,        
        concentration_type = None,
        concentration_value = None,
        concentration_unit_type = None,
        value_date = None,
        value_integer = None,
        value_string = None,
        value_float = None,
        value_datatype = "xsd:date",
        comments = None,
        startdate = None,
        enddate = None,
        age = None,
        uniqid = None,
        event_id = None,
        value = None,
        valueIRI = None,
        activity = None,
        target = None,
        agent = None,
        input = None,
        unit = None
        ),
        
    First_visit = dict(
        pid = None,
        process_type = "http://purl.obolibrary.org/obo/NCIT_C142470",
        attribute_type = "http://purl.obolibrary.org/obo/NCIT_C164021",
        agent_type = None,
        agent_id = None,
        input_type = None,
        #input_id = None,
        target_type = None,
        target_id = None,     
        output_type = "http://purl.obolibrary.org/obo/NCIT_C70856", 
        output_id = None,
        unit_type = None,
        specific_process_type = None,
        protocol_type = None,        
        protocol_id = None, 
        substance_type = None,      
        substance_id = None,
        activity_type = None,
        frequency_type = None,
        frequency_value = None,        
        concentration_type = None,
        concentration_value = None,
        concentration_unit_type = None,
        value_date = None,
        value_integer = None,
        value_string = None,
        value_float = None,
        value_datatype = "xsd:date",
        comments = None,
        startdate = None,
        enddate = None,
        age = None,
        uniqid = None,
        event_id = None,
        value = None,
        valueIRI = None,
        activity = None,
        target = None,
        agent = None,
        input = None,
        unit = None
        ),
        
    Symptoms_onset = dict(
        pid = None,
        process_type = "http://purl.obolibrary.org/obo/NCIT_C142470",
        attribute_type = "http://purl.obolibrary.org/obo/NCIT_C124353",
        agent_type = None,
        agent_id = None,
        input_type = None,
        #input_id = None,
        target_type = None,
        target_id = None,     
        output_type = "http://purl.obolibrary.org/obo/NCIT_C70856", 
        output_id = None,
        unit_type = None,
        specific_process_type = None,
        protocol_type = None,        
        protocol_id = None, 
        substance_type = None,      
        substance_id = None,
        activity_type = None,
        frequency_type = None,
        frequency_value = None,        
        concentration_type = None,
        concentration_value = None,
        concentration_unit_type = None,
        value_date = None,
        value_integer = None,
        value_string = None,
        value_float = None,
        value_datatype = "xsd:date",
        comments = None,
        startdate = None,
        enddate = None,
        age = None,
        uniqid = None,
        event_id = None,
        value = None,
        valueIRI = None,
        activity = None,
        target = None,
        agent = None,
        input = None,
        unit = None
        ),
    
    Sex = dict(
        pid = None,
        process_type = "http://purl.obolibrary.org/obo/NCIT_C142470",
        attribute_type = None,
        agent_type = None,
        agent_id = None,
        input_type = None,
        #input_id = None,
        target_type = None,
        target_id = None,     
        output_type = "http://purl.obolibrary.org/obo/NCIT_C160908", 
        output_id = None,
        unit_type = None,
        specific_process_type = None,
        protocol_type = None,        
        protocol_id = None, 
        substance_type = None,      
        substance_id = None,
        activity_type = None,
        frequency_type = None,
        frequency_value = None,        
        concentration_type = None,
        concentration_value = None,
        concentration_unit_type = None,
        value_date = None,
        value_integer = None,
        value_string = None,
        value_float = None,
        value_datatype = "xsd:string",
        comments = None,
        startdate = None,
        enddate = None,
        age = None,
        uniqid = None,
        event_id = None,
        value = None,
        valueIRI = None,
        activity = None,
        target = None,
        agent = None,
        input = None,
        unit = None
        ),
        
        
    Status = dict(
        pid = None,
        process_type = "http://purl.obolibrary.org/obo/NCIT_C142470",
        attribute_type = None,
        agent_type = None,
        agent_id = None,
        input_type = None,
        #input_id = None,
        target_type = None,
        target_id = None,     
        output_type = "http://purl.obolibrary.org/obo/NCIT_C164628", 
        output_id = None,
        unit_type = None,
        specific_process_type = None,
        protocol_type = None,        
        protocol_id = None, 
        substance_type = None,      
        substance_id = None,
        activity_type = None,
        frequency_type = None,
        frequency_value = None,        
        concentration_type = None,
        concentration_value = None,
        concentration_unit_type = None,
        value_date = None,
        value_integer = None,
        value_string = None,
        value_float = None,
        value_datatype = "xsd:string",
        comments = None,
        startdate = None,
        enddate = None,
        age = None,
        uniqid = None,
        event_id = None,
        value = None,
        valueIRI = None,
        activity = None,
        target = None,
        agent = None,
        input = None,
        unit = None
        ),
    
    Education = dict(
        pid = None,
        process_type = "http://purl.obolibrary.org/obo/NCIT_C142470",
        attribute_type = "http://purl.obolibrary.org/obo/NCIT_C17953",
        agent_type = None,
        agent_id = None,
        input_type = None,
        #input_id = None,
        target_type = None,
        target_id = None,     
        output_type = "http://purl.obolibrary.org/obo/NCIT_C70856", 
        output_id = None,
        unit_type = None,
        specific_process_type = None,
        protocol_type = "http://purl.obolibrary.org/obo/NCIT_C73537",        
        protocol_id = "https://www.wikidata.org/wiki/Q1390351", 
        substance_type = None,      
        substance_id = None,
        activity_type = None,
        frequency_type = None,
        frequency_value = None,        
        concentration_type = None,
        concentration_value = None,
        concentration_unit_type = None,
        value_date = None,
        value_integer = None,
        value_string = None,
        value_float = None,
        value_datatype = "xsd:integer",
        comments = None,
        startdate = None,
        enddate = None,
        age = None,
        uniqid = None,
        event_id = None,
        value = None,
        valueIRI = None,
        activity = None,
        target = None,
        agent = None,
        input = None,
        unit = None
        ),
    
        
        
    Diagnosis = dict(
        pid = None,
        process_type = "http://purl.obolibrary.org/obo/NCIT_C18020",
        attribute_type = None,
        agent_type = None,
        agent_id = None,
        input_type = None,
        #input_id = None,
        target_type = None,
        target_id = None,     
        output_type = "http://purl.obolibrary.org/obo/NCIT_C154625", 
        output_id = None,
        unit_type = None,
        specific_process_type = None,
        protocol_type = None,        
        protocol_id = None, 
        substance_type = None,      
        substance_id = None,
        activity_type = None,
        frequency_type = None,
        frequency_value = None,        
        concentration_type = None,
        concentration_value = None,
        concentration_unit_type = None,
        value_date = None,
        value_integer = None,
        value_string = None,
        value_float = None,
        value_datatype = "xsd:string",
        comments = None,
        startdate = None,
        enddate = None,
        age = None,
        uniqid = None,
        event_id = None,
        value = None,
        valueIRI = None,
        activity = None,
        target = None,
        agent = None,
        input = None,
        unit = None
        ),
        
        
    Phenotype = dict(
        pid = None,
        process_type = "http://purl.obolibrary.org/obo/NCIT_C18020",
        attribute_type = None,
        agent_type = None,
        agent_id = None,
        input_type = None,
        #input_id = None,
        target_type = None,
        target_id = None,     
        output_type = "http://purl.obolibrary.org/obo/NCIT_C16977", 
        output_id = None,
        unit_type = None,
        specific_process_type = None,
        protocol_type = None,        
        protocol_id = None, 
        substance_type = None,      
        substance_id = None,
        activity_type = None,
        frequency_type = None,
        frequency_value = None,        
        concentration_type = None,
        concentration_value = None,
        concentration_unit_type = None,
        value_date = None,
        value_integer = None,
        value_string = None,
        value_float = None,
        value_datatype = "xsd:string",
        comments = None,
        startdate = None,
        enddate = None,
        age = None,
        uniqid = None,
        event_id = None,
        value = None,
        valueIRI = None,
        activity = None,
        target = None,
        agent = None,
        input = None,
        unit = None
        ),
        
        
    Genotype = dict(
        pid = None,
        process_type = "http://purl.obolibrary.org/obo/NCIT_C15709",
        attribute_type = None,
        agent_type = None,
        agent_id = None,
        input_type = None,
        #input_id = None,
        target_type = None,
        target_id = None,     
        output_type = None, 
        output_id = None,
        unit_type = None,
        specific_process_type = None,
        protocol_type = None,        
        protocol_id = None, 
        substance_type = None,      
        substance_id = None,
        activity_type = None,
        frequency_type = None,
        frequency_value = None,        
        concentration_type = None,
        concentration_value = None,
        concentration_unit_type = None,
        value_date = None,
        value_integer = None,
        value_string = None,
        value_float = None,
        value_datatype = "xsd:string",
        comments = None,
        startdate = None,
        enddate = None,
        age = None,
        uniqid = None,
        event_id = None,
        value = None,
        valueIRI = None,
        activity = None,
        target = None,
        agent = None,
        input = None,
        unit = None
        ),
        
        
    # Zygosity = dict(
    #     pid = None,
    #     process_type = "http://purl.obolibrary.org/obo/NCIT_C15709",
    #     attribute_type = "http://purl.obolibrary.org/obo/NCIT_C181350",
    #     agent_type = None,
    #     agent_id = None,
    #     input_type = None,
    #     #input_id = None,
    #     target_type = "http://purl.obolibrary.org/obo/NCIT_C45766",
    #     target_id = None,     
    #     output_type = None, 
    #     output_id = None,
    #     unit_type = None,
    #     specific_process_type = None,
    #     protocol_type = None,        
    #     protocol_id = None, 
    #     substance_type = None,      
    #     substance_id = None,
    #     activity_type = None,
    #     frequency_type = None,
    #     frequency_value = None,        
    #     concentration_type = None,
    #     concentration_value = None,
    #     concentration_unit_type = None,
    #     value_date = None,
    #     value_integer = None,
    #     value_string = None,
    #     value_float = None,
    #     value_datatype = "xsd:string",
    #     comments = None,
    #     startdate = None,
    #     enddate = None,
    #     age = None,
    #     uniqid = None,
    #     event_id = None,
    #     value = None,
    #     valueIRI = None,
    #     activity = None,
    #     target = None,
    #     agent = None,
    #     input = None,
    #     unit = None
    #     ),
        
        
    # Protein = dict(
    #     pid = None,
    #     process_type = "http://purl.obolibrary.org/obo/NCIT_C18962",
    #     attribute_type = None,
    #     agent_type = None,
    #     agent_id = None,
    #     input_type = None,
    #     #input_id = None,
    #     target_type = "http://purl.obolibrary.org/obo/NCIT_C165059",
    #     target_id = None,     
    #     output_type = None, 
    #     output_id = None,
    #     unit_type = None,
    #     specific_process_type = None,
    #     protocol_type = None,        
    #     protocol_id = None, 
    #     substance_type = None,      
    #     substance_id = None,
    #     activity_type = None,
    #     frequency_type = None,
    #     frequency_value = None,        
    #     concentration_type = None,
    #     concentration_value = None,
    #     concentration_unit_type = None,
    #     value_date = None,
    #     value_integer = None,
    #     value_string = None,
    #     value_float = None,
    #     value_datatype = "xsd:string",
    #     comments = None,
    #     startdate = None,
    #     enddate = None,
    #     age = None,
    #     uniqid = None,
    #     event_id = None,
    #     value = None,
    #     valueIRI = None,
    #     activity = None,
    #     target = None,
    #     agent = None,
    #     input = None,
    #     unit = None
    #     ),
        
        
    Consent_contacted = dict(
        pid = None,
        process_type = "http://purl.obolibrary.org/obo/OBI_0000810",
        attribute_type = None,
        agent_type = None,
        agent_id = None,
        input_type = None,
        #input_id = None,
        target_type = None,
        target_id = None,     
        output_type = "http://purl.obolibrary.org/obo/OBIB_0000488", 
        output_id = None,
        unit_type = None,
        specific_process_type = None,
        protocol_type = None,        
        protocol_id = None, 
        substance_type = None,      
        substance_id = None,
        activity_type = None,
        frequency_type = None,
        frequency_value = None,        
        concentration_type = None,
        concentration_value = None,
        concentration_unit_type = None,
        value_date = None,
        value_integer = None,
        value_string = None,
        value_float = None,
        value_datatype = "xsd:string",
        comments = None,
        startdate = None,
        enddate = None,
        age = None,
        uniqid = None,
        event_id = None,
        value = None,
        valueIRI = None,
        activity = None,
        target = None,
        agent = None,
        input = None,
        unit = None
        ),
        
        
    Consent_used = dict(
        pid = None,
        process_type = "http://purl.obolibrary.org/obo/OBI_0000810",
        attribute_type = None,
        agent_type = None,
        agent_id = None,
        input_type = None,
        #input_id = None,
        target_type = None,
        target_id = None,     
        output_type = "http://purl.obolibrary.org/obo/DUO_0000001", 
        output_id = None,
        unit_type = None,
        specific_process_type = None,
        protocol_type = None,        
        protocol_id = None, 
        substance_type = None,      
        substance_id = None,
        activity_type = None,
        frequency_type = None,
        frequency_value = None,        
        concentration_type = None,
        concentration_value = None,
        concentration_unit_type = None,
        value_date = None,
        value_integer = None,
        value_string = None,
        value_float = None,
        value_datatype = "xsd:string",
        comments = None,
        startdate = None,
        enddate = None,
        age = None,
        uniqid = None,
        event_id = None,
        value = None,
        valueIRI = None,
        activity = None,
        target = None,
        agent = None,
        input = None,
        unit = None
        ),
        
        
    Biobank = dict(
        pid = None,
        process_type = "http://purl.obolibrary.org/obo/OMIABIS_0000061",
        attribute_type = None,
        agent_type = "http://purl.obolibrary.org/obo/OBIB_0000616",
        agent_id = None,
        input_type = None,
        #input_id = None,
        target_type = None,
        target_id = None,     
        output_type = "http://purl.obolibrary.org/obo/NCIT_C115570", 
        output_id = None,
        unit_type = None,
        specific_process_type = None,
        protocol_type = None,        
        protocol_id = None, 
        substance_type = None,      
        substance_id = None,
        activity_type = None,
        frequency_type = None,
        frequency_value = None,        
        concentration_type = None,
        concentration_value = None,
        concentration_unit_type = None,
        value_date = None,
        value_integer = None,
        value_string = None,
        value_float = None,
        value_datatype = "xsd:string",
        comments = None,
        startdate = None,
        enddate = None,
        age = None,
        uniqid = None,
        event_id = None,
        value = None,
        valueIRI = None,
        activity = None,
        target = None,
        agent = None,
        input = None,
        unit = None
        ),
    
    # Exemplar = dict(
    #     pid = None,
    #     process_type = None,
    #     attribute_type = None,
    #     agent_type = None,
    #     agent_id = None,
    #     input_type = None,
    #     #input_id = None,
    #     target_type = None,
    #     target_id = None,     
    #     output_type = None, 
    #     output_id = None,
    #     unit_type = None,
    #     specific_process_type = None,
    #     protocol_type = None,        
    #     protocol_id = None, 
    #     substance_type = None,      
    #     substance_id = None,
    #     activity_type = None,
    #     frequency_type = None,
    #     frequency_value = None,        
    #     concentration_type = None,
    #     concentration_value = None,
    #     concentration_unit_type = None,
    #     value_date = None,
    #     value_integer = None,
    #     value_string = None,
    #     value_float = None,
    #     value_datatype = "xsd:string",
    #     comments = None,
    #     startdate = None,
    #     enddate = None,
    #     age = None,
    #     uniqid = None,
    #     event_id = None,
    #     value = None,
    #     valueIRI = None,
    #     activity = None,
    #     target = None,
    #     agent = None,
    #     input = None,
    #     unit = None
    #     ),
    

    Disability = dict(
        pid = None,
        process_type = "http://purl.obolibrary.org/obo/NCIT_C91102",
        attribute_type = "http://purl.obolibrary.org/obo/NCIT_C21007",
        agent_type = None,
        agent_id = None,
        input_type = None,
        #input_id = None,
        target_type = None,
        target_id = None,     
        output_type = "http://purl.obolibrary.org/obo/NCIT_C70856", 
        output_id = None,
        unit_type = None,
        specific_process_type = None,
        protocol_type = None,        
        protocol_id = None, 
        substance_type = None,      
        substance_id = None,
        activity_type = None,
        frequency_type = None,
        frequency_value = None,        
        concentration_type = None,
        concentration_value = None,
        concentration_unit_type = None,
        value_date = None,
        value_integer = None,
        value_string = None,
        value_float = None,
        value_datatype = "xsd:float",
        comments = None,
        startdate = None,
        enddate = None,
        age = None,
        uniqid = None,
        event_id = None,
        value = None,
        valueIRI = None,
        activity = None,
        target = None,
        agent = None,
        input = None,
        unit = None
        ),
    
    Body_measurement = dict(
        pid = None,
        process_type = "http://purl.obolibrary.org/obo/NCIT_C142470",
        attribute_type = None,
        agent_type = None,
        agent_id = None,
        input_type = None,
        #input_id = None,
        target_type = None,
        target_id = None,     
        output_type = "http://purl.obolibrary.org/obo/NCIT_C70856", 
        output_id = None,
        unit_type = None,
        specific_process_type = None,
        protocol_type = None,        
        protocol_id = None, 
        substance_type = None,      
        substance_id = None,
        activity_type = None,
        frequency_type = None,
        frequency_value = None,        
        concentration_type = None,
        concentration_value = None,
        concentration_unit_type = None,
        value_date = None,
        value_integer = None,
        value_string = None,
        value_float = None,
        value_datatype = "xsd:float",
        comments = None,
        startdate = None,
        enddate = None,
        age = None,
        uniqid = None,
        event_id = None,
        value = None,
        valueIRI = None,
        activity = None,
        target = None,
        agent = None,
        input = None,
        unit = None
        ),
    
    Lab_measurement = dict(
        pid = None,
        process_type = "http://purl.obolibrary.org/obo/NCIT_C25294",
        attribute_type = None,
        agent_type = None,
        agent_id = None,
        input_type = None,
        #input_id = None,
        target_type = None,
        target_id = None,     
        output_type = "http://purl.obolibrary.org/obo/NCIT_C70856", 
        output_id = None,
        unit_type = None,
        specific_process_type = None,
        protocol_type = None,        
        protocol_id = None, 
        substance_type = None,      
        substance_id = None,
        activity_type = None,
        frequency_type = None,
        frequency_value = None,        
        concentration_type = None,
        concentration_value = None,
        concentration_unit_type = None,
        value_date = None,
        value_integer = None,
        value_string = None,
        value_float = None,
        value_datatype = "xsd:float",
        comments = None,
        startdate = None,
        enddate = None,
        age = None,
        uniqid = None,
        event_id = None,
        value = None,
        valueIRI = None,
        activity = None,
        target = None,
        agent = None,
        input = None,
        unit = None
        ),
    
    Imaging = dict(
        pid = None,
        process_type = "http://purl.obolibrary.org/obo/NCIT_C17369",
        attribute_type = None,
        agent_type = None,
        agent_id = None,
        input_type = None,
        #input_id = None,
        target_type = None,
        target_id = None,     
        output_type = "http://purl.obolibrary.org/obo/NCIT_C81289", 
        output_id = None,
        unit_type = None,
        specific_process_type = None,
        protocol_type = None,        
        protocol_id = None, 
        substance_type = None,      
        substance_id = None,
        activity_type = None,
        frequency_type = None,
        frequency_value = None,        
        concentration_type = None,
        concentration_value = None,
        concentration_unit_type = None,
        value_date = None,
        value_integer = None,
        value_string = None,
        value_float = None,
        value_datatype = "xsd:string",
        comments = None,
        startdate = None,
        enddate = None,
        age = None,
        uniqid = None,
        event_id = None,
        value = None,
        valueIRI = None,
        activity = None,
        target = None,
        agent = None,
        input = None,
        unit = None
        ),
    
    Surgery = dict(
        pid = None,
        process_type = "http://purl.obolibrary.org/obo/NCIT_C15329",
        attribute_type = None,
        agent_type = None,
        agent_id = None,
        input_type = None,
        #input_id = None,
        target_type = None,
        target_id = None,     
        output_type = None, 
        output_id = None,
        unit_type = None,
        specific_process_type = None,
        protocol_type = "http://purl.obolibrary.org/obo/NCIT_C73537",        
        protocol_id = None, 
        substance_type = None,      
        substance_id = None,
        activity_type = None,
        frequency_type = None,
        frequency_value = None,        
        concentration_type = None,
        concentration_value = None,
        concentration_unit_type = None,
        value_date = None,
        value_integer = None,
        value_string = None,
        value_float = None,
        value_datatype = "xsd:string",
        comments = None,
        startdate = None,
        enddate = None,
        age = None,
        uniqid = None,
        event_id = None,
        value = None,
        valueIRI = None,
        activity = None,
        target = None,
        agent = None
    ),
    
    Medication = dict(
        pid = None,
        process_type = "http://purl.obolibrary.org/obo/NCIT_C70962",
        attribute_type = None,
        agent_type = None,
        agent_id = None,
        input_type = None,
        #input_id = None,
        target_type = None,
        target_id = None,     
        output_type = None, 
        output_id = None,
        unit_type = None,
        specific_process_type = None,
        protocol_type = "http://purl.obolibrary.org/obo/NCIT_C73537",        
        protocol_id = None, 
        substance_type = "http://purl.obolibrary.org/obo/NCIT_C177929",      
        substance_id = None,
        activity_type = None,
        frequency_type = None,
        frequency_value = None,        
        concentration_type = "http://purl.obolibrary.org/obo/NCIT_C25488",
        concentration_value = None,
        concentration_unit_type = None,
        value_date = None,
        value_integer = None,
        value_string = None,
        value_float = None,
        value_datatype = "xsd:float",
        comments = None,
        startdate = None,
        enddate = None,
        age = None,
        uniqid = None,
        event_id = None,
        value = None,
        valueIRI = None,
        activity = None,
        target = None,
        agent = None,
        input = None,
        unit = None
        ),
    
    
    Clinical_trial = dict(
        pid = None,
        process_type = "http://purl.obolibrary.org/obo/NCIT_C71104",
        attribute_type = None,
        agent_type = "http://purl.obolibrary.org/obo/NCIT_C16696",
        agent_id = None,
        input_type = None,
        #input_id = None,
        target_type = None,
        target_id = None,     
        output_type = "http://purl.obolibrary.org/obo/NCIT_C115575", 
        output_id = None,
        unit_type = None,
        specific_process_type = None,
        protocol_type = None,        
        protocol_id = None, 
        substance_type = None,      
        substance_id = None,
        activity_type = None,
        frequency_type = None,
        frequency_value = None,        
        concentration_type = None,
        concentration_value = None,
        concentration_unit_type = None,
        value_date = None,
        value_integer = None,
        value_string = None,
        value_float = None,
        value_datatype = "xsd:string",
        comments = None,
        startdate = None,
        enddate = None,
        age = None,
        uniqid = None,
        event_id = None,
        value = None,
        valueIRI = None,
        activity = None,
        target = None,
        agent = None    )

  )


import pandas as pd
from perseo.main import milisec
import sys
from .template import Template

class Toolkit:

    def import_your_data_from_csv(self, input_data):
      try:
          data = pd.read_csv(input_data)

          return data
      except FileNotFoundError:
          print(f'File not found: {input_data}')
          return None
      except pd.errors.EmptyDataError:
          print(f'The CSV file is empty: {input_data}')
          return None
      except pd.errors.ParserError:
          print(f'Error parsing the CSV file: {input_data}')
          return None
      
    ## Check the status of the columns
    def check_status_column_names(self, data):
        column_names = ['model', 'pid', 'event_id', 'value', 'age', 'value_datatype', 'valueIRI', 'activity', 'unit', 'input', 'target', 'protocol_id', 'frequency_type', 'frequency_value', 'agent', 'startdate', 'enddate', 'comments']

        check_column_in_data = set(data.columns).issubset(column_names)
        
        if not check_column_in_data:
            sys.exit('Your CSV template contains column names that are not part of the accepted for the Toolkit. Accepted column names: {}'.format(column_names))
            
        for name in column_names:
            if name not in data.columns:
                data[name] = pd.Series(dtype=object)
                
        data = data.where(pd.notnull(data), None)

        return data
        
    ## Add ontological columns
    def add_columns_from_template(self,data):        
        print('Transforming and Validating....')

        data = data.where(pd.notnull(data), None)

        temp = Template.template_model
        final_df = pd.DataFrame()

        for row in data.iterrows():
            milisec_point = milisec()

            # Tag each row with the new of the model 
            new_row = {milisec_point : {'model': row[1]['model']}}

            # Include columns related to ontological terms:
            for cde in temp.items():
                if cde[0] == new_row[milisec_point]['model']:
                    new_row[milisec_point].update(cde[1])

            # Include columns from input CSV table:
            new_row[milisec_point].update({title: val for title, val in row[1].items() if val is not None})

            # Concate rows:
            final_row_df = pd.DataFrame(new_row[milisec_point], index=[1])
            if not final_row_df.empty and not final_row_df.isna().all().all():
                final_df = pd.concat([final_df, final_row_df], ignore_index=True)

        final_df = final_df.reset_index(drop=True)
        final_df = final_df.where(pd.notnull(final_df), None)
        return final_df

      
    def transform_shape_based_on_config(self, configuration, data):
        
        # Import static template for all CDE terms:
        temp = Template.template_model
        
        # Empty objects:
        final_df = pd.DataFrame()
        row_df = {}

        # Iterate each row from data input
        # check each YAML object from configuration file to set the parameters
        for row in data.iterrows():

            for config in configuration.items():

                # Create a unique stamp per new row to about them to colapse:
                milisec_point = milisec()

                row_df.update({milisec_point: {'model':config[1]['cde']}})
                
                # Add YAML template static information
                for cde in temp.items():
                    if cde[0] == row_df[milisec_point]['model']:
                        row_df[milisec_point].update(cde[1])

                # Relate each YAML parameter with original data input
                for element in config[1]['columns'].items():
                    for r in row[1].index:
                        if r == element[1]:
                            dict_element = {element[0]:row[1][r]}
                            row_df[milisec_point].update(dict_element)
                            
                # Store formed element into the final table:
                final_row_df = pd.DataFrame(row_df[milisec_point], index=[1])
                if not final_row_df.empty and not final_row_df.isna().all().all():
                    final_df = pd.concat([final_df, final_row_df], ignore_index=True)
        final_df = final_df.reset_index(drop=True)

        final_df = final_df.where(pd.notnull(final_df), None)
        return final_df

    ## Value edition
    def value_edition(self, data):
        not_null_data = pd.notnull(data)

        for index, row in data.iterrows():
            if not_null_data.loc[index, 'value']:
                value = data.loc[index, 'value']
                if row['value_datatype'] == 'xsd:string':
                    data.at[index, 'value_string'] = value
                elif row['value_datatype'] == 'xsd:float' and row['model'] not in ['Medication']:
                    data.at[index, 'value_float'] = value
                elif row['value_datatype'] == 'xsd:integer' and row['model'] not in ['Medication']:
                    data.at[index, 'value_integer'] = value
                elif row['value_datatype'] == 'xsd:date':
                    data.at[index, 'value_date'] = value
                elif row['model'] in ['Medication']:
                    data.at[index, 'concentration_value'] = value

            if 'valueIRI' in data.columns and not_null_data.loc[index, 'valueIRI']:
                value_iri = data.loc[index, 'valueIRI']
                
                if row['model'] in ['Sex', 'Status', 'Diagnosis', 'Phenotype', 'Clinical_trial', 'Body_measurement']:
                    data.at[index, 'attribute_type'] = value_iri
                    
                elif row['model'] in ['Imaging']:
                    data.at[index, 'output_id'] = value_iri
                    
                elif row['model'] in ['Genotype']:
                    data.at[index, 'output_type'] = value_iri

            if 'target' in data.columns and not_null_data.loc[index, 'target']:
                target = data.loc[index, 'target']
                    
                if row['model'] in ['Symptoms_onset', 'Lab_measurement', 'Surgical', 'Imaging', 'Questionnaire', 'Genotype']:
                    data.at[index, 'target_type'] = target

            if 'input' in data.columns and not_null_data.loc[index, 'input']:
                input_value = data.loc[index, 'input']
                
                if row['model'] in ['Genotype', 'Lab_measurement', 'Imaging', 'Biobank']:
                    data.at[index, 'input_type'] = input_value

            if 'agent' in data.columns and not_null_data.loc[index, 'agent']:
                agent = data.loc[index, 'agent']
                
                if row['model'] in ['Biobank', 'Clinical_trial']:
                    data.at[index, 'agent_id'] = agent
                    
                elif row['model'] in ['Medication', 'Surgery']:
                    data.at[index, 'substance_id'] = agent
                
                elif row['model'] in ['Genotype']:
                    data.at[index, 'attribute_type'] = agent

            if 'activity' in data.columns and not_null_data.loc[index, 'activity']:
                activity = data.loc[index, 'activity']
                
                if row['model'] in ['Disability', 'Lab_measurement', 'Imaging', 'Surgery', 'Genotype']:
                    data.at[index, 'specific_process_type'] = activity
                    
                elif row['model'] in ['Medication']:
                    data.at[index, 'activity_type'] = activity

            if 'unit' in data.columns and not_null_data.loc[index, 'unit']:
                unit_value = data.loc[index, 'unit']
                
                if row['model'] in ['Body_measurement', 'Lab_measurement']:
                    data.at[index, 'unit_type'] = unit_value
                    
                elif row['model'] in ['Medication']:
                    data.at[index, 'concentration_unit_type'] = unit_value
                    
        return data   

    # Time edition
    def time_edition(self, data):
        
        data = data.where(pd.notnull(data), None)

        for index, row in data.iterrows():
            ## From startdate to enddate
            
            if type(row['enddate']) == float or row['enddate'] == None: #TODO work on nan values to filter them better
                data.at[index, 'enddate'] = row['startdate']
                
        return data

    #Clean rows with no value
    
    def clean_empty_rows(self, data):
        
        data = data.where(pd.notnull(data), None)
        columns_to_check = ['value', 'valueIRI', 'activity', 'target', 'agent']

        for row in data.iterrows():
            if all(row[1][col] is None for col in columns_to_check):
                if row[1]['model'] not in ['Biobank', 'Consent_used', 'Consent_contacted']: 
                    data = data.drop(row[0])
        return data
    
    def delete_extra_columns(self, data):
        columns_to_delete = ['value', 'valueIRI', 'target', 'agent', 'input', 'activity', 'unit']
        
        for column in columns_to_delete:
            if column in data.columns:
                del data[column]

        return data
    
    def unique_id_generation(self,data):
        data['uniqid'] = ''

        for i in data.index:
            data.at[i, 'uniqid'] = milisec()
        
        return data

    def whole_quality_control(self,input_data):

        imported_file = self.import_your_data_from_csv(input_data)
        
        if imported_file is not None:
            print('CSV file imported successfully.')
        else:
            print('CSV file import failed. Please check the file path and format.')

        columns_names_conformation = self.check_status_column_names(imported_file)
        
        table_with_template_addition = self.add_columns_from_template(columns_names_conformation)

        table_with_value_edited = self.value_edition(table_with_template_addition)

        table_with_time_corrected = self.time_edition(table_with_value_edited)

        table_with_blanks_cleaned = self.clean_empty_rows(table_with_time_corrected)

        table_extra_column_deleted = self.delete_extra_columns(table_with_blanks_cleaned)

        table_with_uniqid = self.unique_id_generation(table_extra_column_deleted)

        if table_with_uniqid is not None:
            print('CSV data transformation done.')
        else:
            sys.exit('CSV file quiality control failed. Please check the columns names, every required column is not present')

        return table_with_uniqid
      
      
    def yaml_quality_control(self,input_data,configuration):

        imported_file = self.import_your_data_from_csv(input_data)
        if imported_file is not None:
            print('CSV file imported successfully.')
        else:
            print('CSV file import failed. Please check the file path and format.')

        # columns_names_conformation = self.check_status_column_names(imported_file)
        # if columns_names_conformation is not None:
        #     print('Every CSV columns present.')
        # else:
        #     sys.exit('CSV file quiality control failed. Please check the columns names, every required column is not present')

        table_with_template_addition = self.transform_shape_based_on_config(configuration=configuration, data=imported_file)

        table_with_value_edited = self.value_edition(table_with_template_addition)

        table_with_time_corrected = self.time_edition(table_with_value_edited)

        table_with_blanks_cleaned = self.clean_empty_rows(table_with_time_corrected)

        table_extra_column_deleted = self.delete_extra_columns(table_with_blanks_cleaned)

        table_with_uniqid = self.unique_id_generation(table_extra_column_deleted)

        if table_with_uniqid is not None:
            print('CSV data transformation done.')
        else:
            print('CSV data transformation failed. Something went wrong during transformation.')

        return table_with_uniqid



from setuptools import setup

setup(
    name="CARE-SM-Toolkit",
    version="0.0.18",
    packages=["toolkit"],
    author="Pablo Alarcón Moreno",
    author_email="pabloalarconmoreno@gmail.com",
    url="https://github.com/CARE-SM/CARE-SM-Toolkit",
    description="CARE-SM Toolkit",
    license="MIT",
    keywords=["EJP-RD","EJP","FAIR-in-a-box","Fiab","CARE-SM"]
)


import pytest
import pandas as pd
from toolkit.main import Toolkit as Tool

@pytest.mark.parametrize("input_data", [
    ("exemplar_data/fiab_input_data.csv")
])
def test_import_data_works(input_data):
    toolkit = Tool()
    result = toolkit.import_your_data_from_csv(input_data)
    assert result is not False

@pytest.mark.parametrize("input_data", [
    (pd.DataFrame({'model': [1, 2], 'pid': [1, 2], 'event_id': [1, 2], 'value': [1, 2], 'age': [1, 2], 'value_datatype': [1, 2], 'valueIRI': [1, 2], 'activity': [1, 2], 'protocol_id':[1, 2],'unit': [1, 2], 'input': [1, 2], 'target': [1, 2], 'intensity': [1, 2],'frequency_type': [1, 2], 'frequency_value': [1, 2], 'agent_id': [1, 2], 'route': [1, 2], 'startdate': [1, 2], 'enddate': [1, 2], 'comments': [1, 2]})),
    (pd.DataFrame({'model': [1, 2], 'pid': [1, 2],'pid2': [1, 2], 'event_id': [1, 2], 'value': [1, 2], 'age': [1, 2], 'value_datatype': [1, 2], 'valueIRI': [1, 2], 'activity': [1, 2],'protocol_id':[1, 2], 'unit': [1, 2], 'input': [1, 2], 'target': [1, 2],'intensity': [1, 2], 'frequency_type': [1, 2], 'frequency_value': [1, 2], 'agent_id': [1, 2], 'route': [1, 2], 'startdate': [1, 2], 'enddate': [1, 2], 'comments': [1, 2]})),
])
def test_column_names_are_present(input_data):
    toolkit = Tool()
    result = toolkit.check_status_column_names(input_data)
    assert result is not False

@pytest.mark.parametrize("input_data", [
    (pd.DataFrame({'model': [1, 2], 'event_id': [1, 2], 'value': [1, 2], 'age': [1, 2], 'value_datatype': [1, 2], 'valueIRI': [1, 2], 'activity': [1, 2],'protocol_id':[1, 2], 'unit': [1, 2], 'input': [1, 2], 'target': [1, 2], 'intensity': [1, 2],'frequency_type': [1, 2], 'frequency_value': [1, 2], 'agent_id': [1, 2], 'route': [1, 2], 'startdate': [1, 2], 'enddate': [1, 2], 'comments': [1, 2]})),
    (pd.DataFrame({'model': [1, 2], 'pid': [1, 2],'pid2': [3, 4], 'value': [1, 2], 'age': [1, 2], 'value_datatype': [1, 2], 'valueIRI': [1, 2], 'activity': [1, 2], 'protocol_id':[1, 2],'unit': [1, 2], 'input': [1, 2], 'target': [1, 2], 'intensity': [1, 2],'frequency_type': [1, 2], 'frequency_value': [1, 2], 'agent_id': [1, 2], 'route': [1, 2], 'startdate': [1, 2], 'enddate': [1, 2], 'comments': [1, 2]})),
])
def test_column_names_are_not_present(input_data):
    toolkit = Tool()
    result = toolkit.check_status_column_names(input_data)
    assert result == False

@pytest.mark.parametrize("data_with_duplicates", [
    (pd.DataFrame({'model': ["model", 2], 'pid': [1, 2],'event_id': [1, 2], 'value': [1, 2], 'age': [1, 2], 'value_datatype': [1, 2], 'valueIRI': [1, 2], 'activity': [1, 2], 'protocol_id':[1, 2], 'unit': [1, 2], 'input': [1, 2], 'target': [1, 2], 'intensity': [1, 2],'frequency_type': [1, 2], 'frequency_value': [1, 2], 'agent_id': [1, 2], 'route': [1, 2], 'startdate': [1, 2], 'enddate': [1, 2], 'comments': [1, 2]})),
    (pd.DataFrame({'model': ["model", "model"], 'pid': [1, 2], 'event_id': [1, 2], 'value': [1, 2], 'age': [1, 2], 'value_datatype': [1, 2], 'valueIRI': [1, 2], 'activity': [1, 2], 'protocol_id':[1, 2], 'unit': [1, 2], 'input': [1, 2], 'target': [1, 2], 'intensity': [1, 2],'frequency_type': [1, 2], 'frequency_value': [1, 2], 'agent_id': [1, 2], 'route': [1, 2], 'startdate': [1, 2], 'enddate': [1, 2], 'comments': [1, 2]})),
])
def test_duplicated_titles_amog_row(data_with_duplicates):
    toolkit = Tool()
    result_with_duplicates = toolkit.check_for_duplicated_titles_among_row(data_with_duplicates)
    assert result_with_duplicates == True

@pytest.mark.parametrize("input_data, expected_data", [

    (pd.DataFrame({'model':['Sex'],'valueIRI':['https://Female'],'value': ["Female"],'target': [None], 'value_datatype': ["xsd:string"], 'value_string': [None],'value_id':[None],'attribute_type':[None]}),
     pd.DataFrame({'model':['Sex'],'valueIRI':['https://Female'],'value': ["Female"],'target': [None], 'value_datatype': ["xsd:string"], 'value_string': ["Female"],'value_id':[None],'attribute_type':['https://Female']})),
    
    (pd.DataFrame({'model':['Birthdate'],'valueIRI':[None],'value': ["2021-12-05"],'target': [None], 'value_datatype': ["xsd:date"], 'value_date': [None],'value_id':[None]}),
     pd.DataFrame({'model':['Birthdate'],'valueIRI':[None],'value': ["2021-12-05"],'target': [None], 'value_datatype': ["xsd:date"], 'value_date': ["2021-12-05"],'value_id':[None]})),
    ])

def test_value_edition_works(input_data, expected_data):
    toolkit = Tool()
    result_value_edition = toolkit.value_edition(input_data)
    
    if "value_integer" in result_value_edition:
        result_value_edition["value_integer"] = result_value_edition["value_integer"].astype('int64')
    if "value_float" in result_value_edition:
        result_value_edition["value_float"] = result_value_edition["value_float"].astype('float')
    if ("value_string" in result_value_edition):
        result_value_edition["value_string"] = result_value_edition["value_string"].astype('object')
    if ("value_date" in result_value_edition):
        result_value_edition["value_date"] = result_value_edition["value_date"].astype('object')
        
    assert result_value_edition.equals(expected_data)

@pytest.mark.parametrize("input_data, expected_data", [
    (pd.DataFrame(
        { 'model': ['Birthdate','Deathdate','Deathdate'],
                  'startdate':[None,'2020-03-08',None],
                   'enddate':[None,None,None],
             }),
     
     pd.DataFrame({ 'model': ['Birthdate','Deathdate','Deathdate'],
                  'startdate':[None,'2020-03-08',None],
                   'enddate':[None,'2020-03-08',None],

             })),

])
def test_time_edition_works(input_data, expected_data):
    toolkit = Tool()
    result_value_edition = toolkit.time_edition(input_data)
    
    if "value_integer" in result_value_edition:
        result_value_edition["age"] = result_value_edition["age"].astype('float64')
        result_value_edition["value_integer"] = result_value_edition["value_integer"].astype('float64')

    # print("Actual DataFrame Data Types:")
    # print(result_value_edition)

    # print("Expected DataFrame Data Types:")
    # print(expected_data)
    assert result_value_edition.equals(expected_data)

@pytest.mark.parametrize("input_data, expected_data", [
    (pd.DataFrame(
        { 'model': ['Diagnosis',"Consent_used","Symptom_onset",'Clinical_trial','Biobank','Biobank'],
                'value':['valueX',None,'valueX',None,None,None],
                'valueIRI':['valueX',None,None,'valueX',None,None],
                'age':['valueX',None,None,None,None,None],
                'agent_id':[None,None,None,'valueX',None,None],
                'activity':[None,None,None,None,None,None],
                'target':[None,None,None,None,None,None],
             }),
     
    pd.DataFrame(
        { 'model': ['Diagnosis',"Consent_used","Symptom_onset",'Clinical_trial','Biobank','Biobank'],
                'value':['valueX',None,'valueX',None,None,None],
                'valueIRI':['valueX',None,None,'valueX',None,None],
                'age':['valueX',None,None,None,None,None],
                'agent_id':[None,None,None,'valueX',None,None],
                'activity':[None,None,None,None,None,None],
                'target':[None,None,None,None,None,None],

             })
    ),
])
def test_clean_empty_rows_works(input_data, expected_data):
    toolkit = Tool()
    result_value_edition = toolkit.clean_empty_rows(input_data)
    assert result_value_edition.equals(expected_data)

if __name__ == '__main__':
    pytest.main([__file__])